/* ========================================================================== */
/*   Clock.c                                                                  */
/*   (c) 2007 St                                                              */
/*                                                                            */
/*   Program for nixie clock controlled ATtiny2313                            */
/*   Xtall frequency: 12 MHz                                                  */
/* ========================================================================== */

// Warning! Don't use bit manipulation suggested in WinAVR examples!
// Setting bit e.g. PORTB |= _BV(PB7) this way can cause problem!
// Value from port is read at first, modificated and written at last.
// Interrupt (just right after read port value) working with the same 
// port can modificate port value. Following write will destroy this data
// written by ISR (interrupt-service-routine). Use direct bit manipulation
// instead (asm instructions "sbi" and "cbi").

#include <avr/io.h>
#include <avr/interrupt.h>

#define word unsigned int
#define byte unsigned char

// Definition of PIN connection
#define POSITION0_0 asm ("cbi 0x12,4" ::)  // digit anode switches
#define POSITION0_1 asm ("sbi 0x12,4" ::)
#define POSITION1_0 asm ("cbi 0x12,3" ::)
#define POSITION1_1 asm ("sbi 0x12,3" ::)
#define POSITION2_0 asm ("cbi 0x12,2" ::)
#define POSITION2_1 asm ("sbi 0x12,2" ::)
#define POSITION3_0 asm ("cbi 0x12,1" ::)
#define POSITION3_1 asm ("sbi 0x12,1" ::)
#define POSITION4_0 asm ("cbi 0x12,0" ::)
#define POSITION4_1 asm ("sbi 0x12,0" ::)
#define POSITION5_0 asm ("cbi 0x18,6" ::)
#define POSITION5_1 asm ("sbi 0x18,6" ::)

#define DIGIT0_0 asm ("cbi 0x18,5" ::)	// outputs for nixie driver 74141
#define DIGIT0_1 asm ("sbi 0x18,5" ::)
#define DIGIT1_0 asm ("cbi 0x18,2" ::)
#define DIGIT1_1 asm ("sbi 0x18,2" ::)
#define DIGIT2_0 asm ("cbi 0x18,0" ::)
#define DIGIT2_1 asm ("sbi 0x18,0" ::)
#define DIGIT3_0 asm ("cbi 0x18,4" ::)
#define DIGIT3_1 asm ("sbi 0x18,4" ::)

#define BUZZER_0 asm ("cbi 0x18,7" ::)	// alarm buzzer
#define BUZZER_1 asm ("sbi 0x18,7" ::)
#define KEYINP   PINB & 0x02			// button input

// Connection of function buttons to digit positions
#define KEYMENU  4 
#define KEYUP    3 
#define KEYENTER 2 

// Global variables used in main loop and in the timer interrupt
// counter for 250 microseconds ticks
word milisec = 0;
// arrays for real time, alarm and auxiliary time during setting
byte time[6],alarm[6],auxtime[6];
// pressed button code, half second flag (for alarm and blink), blinking position 
byte button = 0, halfsec = 0, blink = 0xF;
// pointer which array to display
byte *p_digit = time;

// wait for button pressed or 20 second time-out
void waitbutton(void) {
  word i;
  byte n;
  
  for(n=0; n<200; n++) {
    // Wait for about 100 msec
    for(i=0; i<60000; i++) {} 
    if(button) break;
  }
}

// wait for all buttons released
void endbutton(void) {
  word i;

  while(button) {
    button = 0;
    for(i=0; i<10000; i++) {} 
  }
}

// function for setting array "setdata" with a new time
// seconds are not updated, just only hours and minutes
byte settime(byte *setdata) {
  byte posit = 2, updated = 0, next = 1;

  while(next) {
    blink = posit;   // updated position blinks
    waitbutton();
    // Test buttons and do 1 of 4 possibilities
    switch(button) {
      case KEYMENU : next = 0; break; // Stop updating time
      case KEYENTER: {  // Move to next digit
        posit++;
        if(posit == 6) posit = 2;
      } break;
      case KEYUP   : {  // Increment digit on this position
        setdata[posit]++;
        if(((posit == 2) || (posit == 4)) && (setdata[posit] == 10)) setdata[posit]=0;
        if((posit == 3) && (setdata[posit] == 6)) setdata[posit]=0;
        if((posit == 5) && (setdata[posit] == 3)) setdata[posit]=0;
        updated = 1;
      } break;
      default      : next = 0;  // Time-out expired
    }
    endbutton();
  }
  // Hours above 23 are not allowed
  if((setdata[5] == 2) && (setdata[4] > 3)) {
    setdata[5] = 0; setdata[4] = 0;
  }
  blink = 0xF;
  return(updated);
}
// ****************************
//    Main program function  //
// ****************************
int main(void) {

  // Default values for start
  time[4] = 7; time[3] = 3;
  alarm[4] = 7; alarm[3] = 4;
  alarm[1] = 0xF; alarm[0] = 0xF;

  /* Enable PORTB as output (except PB1)*/
  DDRB = 0xFD;
  /* Enable PORTD as output*/
  DDRD = 0xFF;
  /* Disable pull-ups*/
  MCUCR |= 0x80;

  // Initiative setting Timer 1
  // TOP for timer set to 11999 counts
  OCR1A = 0x2EDF;
  /* Timer 1 is 1 milisec CTC timer */
  TCCR1A = 0x00;
  TCCR1B = 0x9;
  /* Enable timer 1 overflow interrupt. */
  TIMSK = _BV (OCIE1A);
  sei ();

  /***** Main loop of program *****/
  for(;;) {
    // Set alarm Buzzer if time == alarm time
    if((time[5] == alarm[5]) && (time[4] == alarm[4]) && (time[3] == alarm[3]) && (time[2] == alarm[2]) && halfsec) BUZZER_0; else BUZZER_1;
    
    if(button == KEYMENU) {
    // Menu button enters time updating
      endbutton();
      p_digit = alarm;
      if(!settime(alarm)) {
        // Alarm was not updated, try Time
        cli();
        // We must not destroy real Time, copy it to auxiliary array
        auxtime[5] = time[5];auxtime[4] = time[4];auxtime[3] = time[3];auxtime[2] = time[2];
        sei();
        p_digit = auxtime;
        if(settime(auxtime)) {
          // Time was updated - copy it back
          cli();
          time[5] = auxtime[5];time[4] = auxtime[4];time[3] = auxtime[3];
          time[2] = auxtime[2];time[1] = 0;time[0] = 0;
          sei();
        }
      }
      p_digit = time;
    } else button = 0;
      
  } /* loop forever */
  /* please make sure that you never leave main */
  return (0);
}

/*
** ===================================================================
**     Interrupt handler
** ===================================================================
*/
ISR (TIMER1_COMPA_vect)
{
  static const byte tabledigit[] = {3,2,7,4,1,9,8,0,5,6,0xF,0xF,0xF,0xF,0xF,0xF};
  static byte position = 0;
  static byte auxdigit;

  // Increment Time
  milisec++;
  if(milisec == 500) halfsec = 1;
  if(milisec == 1000) {
    halfsec = 0;                 // half second flag
    milisec = 0;                 // miliseconds
    time[0]++;
    if(time[0] == 10) {          // seconds
      time[0] = 0;
      time[1]++;
      if(time[1] == 6) {
        time[1] = 0;
        time[2]++;
        if(time[2] == 10) {      // minutes
          time[2] = 0;
          time[3]++;
          if(time[3] == 6) {
            time[3] = 0;
            time[4]++;
            if(time[4] == 10) {  // hours
              time[4] = 0;
              time[5]++;
            }
            if((time[4] == 4) && (time[5] == 2)) {
              time[4] = 0;
              time[5] = 0;
            }
          }
        }
      }
    }
  }
  // Clear display (all digits off)
  POSITION0_0; POSITION1_0; POSITION2_0; POSITION3_0; POSITION4_0; POSITION5_0;
  // New digit to display
  // Test for blinking digit (and set it according "blink" variable)
  if((blink == position) && halfsec) {
    DIGIT0_1; DIGIT1_1; DIGIT2_1; DIGIT3_1;
  } else {
    // Set new digit to display
    auxdigit = tabledigit[*(p_digit + position)];
    DIGIT0_0; DIGIT1_0; DIGIT2_0; DIGIT3_0;
    if(auxdigit & 0x1) DIGIT0_1;
    if(auxdigit & 0x2) DIGIT1_1;
    if(auxdigit & 0x4) DIGIT2_1;
    if(auxdigit & 0x8) DIGIT3_1;
  }

  // Switch ON new digit
  switch(position) {
    case 0: POSITION0_1; break;
    case 1: POSITION1_1; break;
    case 2: POSITION2_1; break;
    case 3: POSITION3_1; break;
    case 4: POSITION4_1; break;
    case 5: POSITION5_1; break;
  }

  // Test pressed button with this position  
  if((!button) && (KEYINP)) button = position;
  
  // Update position for next digit
  position++;
  if(position == 6) position = 0;
}

/* END */
 
